from snapgene_reader import snapgene_file_to_dict
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML
import math
import json
import random

import os

import pandas as pd
from Bio import SeqIO

# dot_dna_file_folder is the folder where all the plasmids in .dna format reside
# blast_database_name is the name you want to set as the BLAST database
# blast_database_output_folder is the folder where you want the BLAST database files to reside
# most of the time blast_database_name should be the same as $BLASTDB
def snapgene_files_to_blast_database(dot_dna_file_folder, blast_database_name, blast_database_output_folder):
    sub_items_to_dir = os.listdir(dot_dna_file_folder)
    names_for_fasta = []

    fasta_file_contents = ""

    for item in sub_items_to_dir:
        if len(item) >= 4 and item[-4:] == ".dna":
            file_path = "%s/%s" % (dot_dna_file_folder, item)

            name_for_fasta = item.split(".")[0]

            if name_for_fasta in names_for_fasta:
                raise ValueError("%s is repeated at least once as a .dna filename, make the names unique to avoid this error")
            else:
                names_for_fasta.append(name_for_fasta)

            fasta_file_contents += ">%s\n" % (name_for_fasta)

            try:
                snapgene_file_dictionary = snapgene_file_to_dict(file_path)

                sequence = (2 * snapgene_file_dictionary["seq"]).upper()
            except Exception:
                print(file_path)

            fasta_file_contents += sequence + "\n"

    if not os.path.exists(blast_database_output_folder):
        os.makedirs(blast_database_output_folder)

    output_writer = open("%s/%s.fasta" % (blast_database_output_folder, blast_database_name), "w")
    output_writer.write(fasta_file_contents)
    output_writer.close()

    os.system("cd %s" % (blast_database_output_folder))
    os.system("makeblastdb -in %s/%s.fasta -parse_seqids -title \"%s\" -dbtype nucl" % (blast_database_output_folder, blast_database_name, blast_database_name))

def snapgene_files_to_annotation_blast_database_and_csv(dot_dna_file_folder, blast_database_name, blast_database_output_folder):
    sub_items_to_dir = os.listdir(dot_dna_file_folder)

    fasta_file_contents = ""

    parts_df = pd.DataFrame(columns = ["Part Name", "Part Plasmid", "Start Position", "End Position", "Color", "Sequence", "Strand"])

    parts_key_df = pd.DataFrame(columns = ["Part Name FASTA", "Part Name", "Part Plasmid"])

    annotation_number = 1
    for item in sub_items_to_dir:
        file_path = "%s/%s" % (dot_dna_file_folder, item)
        plasmid_name = item.split(".")[0]
        name_for_fasta = "annotation_%s" % (annotation_number)

        if len(item) >= 4 and item[-4:] == ".dna":
            snapgene_file_dictionary = snapgene_file_to_dict(file_path)

            sequence = snapgene_file_dictionary["seq"].upper()

            for feature in snapgene_file_dictionary["features"]:
                feature_position_range = feature["segments"][0]["@range"].split("-")

                start_position = int(feature_position_range[0])
                end_position = int(feature_position_range[1])

                part_sequence = ""
                if start_position < end_position:
                    part_sequence = sequence[start_position - 1: end_position]
                else:
                    part_sequence = sequence[start_position - 1: ] + sequence[0: end_position]

                parts_df = parts_df.append({
                    "Part Name": feature["name"],
                    "Part Plasmid": plasmid_name,
                    "Start Position": start_position,
                    "End Position": end_position,
                    "Color": feature["color"],
                    "Sequence": part_sequence,
                    "Strand": feature["strand"]
                }, ignore_index = True)

                parts_key_df = parts_key_df.append({
                    "Part Name FASTA": name_for_fasta,
                    "Part Name": feature["name"],
                    "Part Plasmid": plasmid_name
                }, ignore_index = True)

                fasta_file_contents += ">%s\n%s\n" % (name_for_fasta, part_sequence)

                annotation_number += 1
                name_for_fasta = "annotation_%s" % (annotation_number)
        elif len(item) >= 3 and item[-3:] == ".gb":
            try:
                for index, record in enumerate(SeqIO.parse(file_path, "genbank")):
                    for feature in record.features:
                        if "label" in feature.qualifiers:
                            annotation_name = feature.qualifiers["label"][0]
                            annotation_location = feature.location

                            annotation_sequence = annotation_location.extract(record.seq)

                            parts_df = parts_df.append({
                                "Part Name": annotation_name,
                                "Part Plasmid": plasmid_name,
                                "Start Position": annotation_location.start,
                                "End Position": annotation_location.end,
                                "Sequence": annotation_sequence,
                                "Strand": annotation_location.strand
                            }, ignore_index = True)

                            parts_key_df = parts_key_df.append({
                                "Part Name FASTA": name_for_fasta,
                                "Part Name": annotation_name,
                                "Part Plasmid": plasmid_name
                            }, ignore_index = True)

                            fasta_file_contents += ">%s\n%s\n" % (name_for_fasta, annotation_sequence)

                            annotation_number += 1
                            name_for_fasta = "annotation_%s" % (annotation_number)
            except UnicodeDecodeError:
                print("Mlem")

    if not os.path.exists(blast_database_output_folder):
        os.makedirs(blast_database_output_folder)

    output_writer = open("%s/%s.fasta" % (blast_database_output_folder, blast_database_name), "w")
    output_writer.write(fasta_file_contents)
    output_writer.close()

    parts_df.to_csv("%s/Parts.csv" % (blast_database_output_folder), index = False)
    parts_key_df.to_csv("%s/Parts Key.csv" % (blast_database_output_folder), index = False)

    os.system("cd %s" % (blast_database_output_folder))
    os.system("makeblastdb -in %s/%s.fasta -parse_seqids -title \"%s\" -dbtype nucl -max_file_sz '2GB'" % (blast_database_output_folder, blast_database_name, blast_database_name))

def conduct_alignment_with_fasta(fasta_file_location, database_name, remove_fasta_file_and_output = False):
    folder_of_fasta_file = os.path.dirname(os.path.abspath(fasta_file_location))
    filename = os.path.basename(fasta_file_location)
    filename_without_extension = os.path.splitext(filename)[0]

    blastx_cline = NcbiblastnCommandline(query = fasta_file_location, db = database_name, out = "%s/%s.xml" % (folder_of_fasta_file, filename_without_extension), outfmt = 5, num_threads = 7)
    blastx_cline()

    xml_file_location = "%s/%s.xml" % (folder_of_fasta_file, filename_without_extension)
    result_handle = open(xml_file_location)

    blast_records = NCBIXML.parse(result_handle)
    num_searches = 0
    response_dict = {}
    while True:
        try:
            blast_record = next(blast_records)
            num_searches += 1
            query_name = blast_record.query

            query_alignment_dict = {}
            num_source_sequences = 0
            for alignment in blast_record.alignments:
                num_source_sequences += 1
                source_sequence_name = alignment.accession
                source_sequence_dict = {}
                num_hsps = 0
                for hsp in alignment.hsps:
                    num_hsps += 1
                    score = hsp.score
                    match = hsp.match
                    source_sequence = hsp.sbjct

                    regions_of_alignment_in_source_sequence = []

                    temporary_interval = []
                    source_sequence_start_position = hsp.sbjct_start - 1
                    position_subtraction = 0

                    alignment_similarity = 0
                    total_positions = 0

                    for i in range(len(source_sequence)):
                        if match[i] == "|":
                            if temporary_interval == []:
                                temporary_interval.append(source_sequence_start_position + i - position_subtraction)
                            if i == len(source_sequence) - 1:
                                temporary_interval.append(source_sequence_start_position + i - position_subtraction)
                                regions_of_alignment_in_source_sequence.append(temporary_interval)
                                temporary_interval = []

                            alignment_similarity += 1
                        else:
                            if temporary_interval != []:
                                temporary_interval.append(source_sequence_start_position + i - 1 - position_subtraction)
                                regions_of_alignment_in_source_sequence.append(temporary_interval)
                                temporary_interval = []
                                if source_sequence[i] == "-":
                                    position_subtraction += 1

                        total_positions += 1

                    #print(regions_of_alignment_in_source_sequence)
                    hsp_dict = {"score": score, "regions_of_alignment": regions_of_alignment_in_source_sequence, "similarity": (alignment_similarity / total_positions) * 100}
                    source_sequence_dict[num_hsps] = hsp_dict

                source_sequence_dict["num_alignments"] = num_hsps
                source_sequence_dict["plasmid_name"] = source_sequence_name
                query_alignment_dict[num_source_sequences] = source_sequence_dict

            query_alignment_dict["num_plasmids"] = num_source_sequences
            query_alignment_dict["query_name"] = query_name
            response_dict[num_searches] = query_alignment_dict
        except StopIteration:
            break

    response_dict["num_searches"] = num_searches
    result_handle.close()

    if remove_fasta_file_and_output:
        os.remove(fasta_file_location)
        os.remove(xml_file_location)

    return response_dict

def conduct_alignment_with_sequence_list(sequences, temp_dir, database_name):
    random_temp_fasta_name = str(random.getrandbits(128))

    while os.path.exists("%s/%s.fasta" % (temp_dir, random_temp_fasta_name)):
        random_temp_fasta_name = str(random.getrandbits(128))

    fasta_file_path = "%s/%s.fasta" % (temp_dir, random_temp_fasta_name)

    fasta_file = open(fasta_file_path, "w")

    for i in range(len(sequences)):
        fasta_file.write(">sequence_%s\n" % (str(i)))
        fasta_file.write("%s\n" % (sequences[i]))

    fasta_file.close()

    return conduct_alignment_with_fasta(fasta_file_path, database_name, remove_fasta_file_and_output = True)

def conduct_alignment_with_snapgene_file_annotations(snapgene_file_folder, temp_dir, database_name, fragment_toggle = False):
    sub_items_to_dir = os.listdir(snapgene_file_folder)

    parts_df = pd.DataFrame(columns = ["Part Name", "Part Plasmid", "Start Position", "End Position", "Color", "Sequence"])

    parts_key_df = pd.DataFrame(columns = ["Part Name FASTA", "Part Name", "Part Plasmid"])

    part_sequences = []

    annotation_number_to_part = {}

    annotation_number = 1
    for item in sub_items_to_dir:
        if len(item) >= 4 and item[-4:] == ".dna":
            file_path = "%s/%s" % (snapgene_file_folder, item)

            plasmid_name = item.split(".")[0]

            snapgene_file_dictionary = snapgene_file_to_dict(file_path)

            sequence = snapgene_file_dictionary["seq"].upper()

            for feature in snapgene_file_dictionary["features"]:
                if fragment_toggle == False or (fragment_toggle == True and ("fragment" in feature["name"] or "Fragment" in feature["name"])):
                    feature_position_range = feature["segments"][0]["@range"].split("-")

                    start_position = int(feature_position_range[0])
                    end_position = int(feature_position_range[1])

                    part_sequence = ""
                    if start_position < end_position:
                        part_sequence = sequence[start_position - 1: end_position]
                    else:
                        part_sequence = sequence[start_position - 1: ] + sequence[0: end_position]

                    part_sequences.append(part_sequence)

                    annotation_number_to_part[annotation_number] = [feature["name"], plasmid_name]

                    annotation_number += 1

    results = conduct_alignment_with_sequence_list(part_sequences, temp_dir, database_name)

    return [results, annotation_number_to_part]
