from snapgene_reader import snapgene_file_to_dict
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML
import math
import json
import random

from iBioFabBackend.models import Plasmid, PlasmidPart
import os

from whoosh import index
from whoosh.fields import Schema, TEXT, BOOLEAN, NUMERIC
from whoosh.qparser import QueryParser

import Bio
from Bio import SeqIO

def refresh_plasmid_and_plasmid_parts_database(dot_dna_file_folder):
    Plasmid.objects.all().delete()
    PlasmidPart.objects.all().delete()

    sub_items_to_dir = os.listdir(dot_dna_file_folder)

    plasmid_schema = Schema(
        name = TEXT(stored = True),
        sequence = TEXT(stored = True),
        sql_id = NUMERIC(stored = True)
    )

    plasmid_parts_schema = Schema(
        name = TEXT(stored = True),
        start_position = NUMERIC(stored = True),
        end_position = NUMERIC(stored = True),
        strand = TEXT(stored = True),
        color = TEXT(stored = True),
        sql_id = NUMERIC(stored = True)
    )

    plasmid_ix = index.create_in("Indexed_Search_Database/Plasmid_Name", plasmid_schema)
    plasmid_writer = plasmid_ix.writer()

    plasmid_parts_ix = index.create_in("Indexed_Search_Database/Plasmid_Annotation", plasmid_parts_schema)
    plasmid_parts_writer = plasmid_parts_ix.writer()

    for item in sub_items_to_dir:
        file_path = "%s/%s" % (dot_dna_file_folder, item)

        if len(item) >= 4 and item[-4:] == ".dna":

            name_for_fasta = item.split(".")[0]
            snapgene_file_dictionary = {}
            try:
                snapgene_file_dictionary = snapgene_file_to_dict(file_path)

                plasmid = None

                if not Plasmid.objects.filter(name = name_for_fasta).exists():
                    plasmid = Plasmid(name = name_for_fasta, sequence = snapgene_file_dictionary["seq"].upper())
                    plasmid.save()
                else:
                    plasmid = Plasmid.objects.get(name = name_for_fasta)

                plasmid_writer.add_document(
                    name = name_for_fasta,
                    sequence = snapgene_file_dictionary["seq"].upper(),
                    sql_id = plasmid.id
                )

                for part in snapgene_file_dictionary["features"]:
                    part_name = part["name"]
                    part_strand = part["strand"]
                    part_color = part["color"]

                    range_str = part["segments"][0]["@range"]
                    range_str_split = range_str.split("-")

                    part_start_position = int(range_str_split[0])
                    part_end_position = int(range_str_split[1])

                    if not PlasmidPart.objects.filter(name = part_name, start_position = part_start_position, end_position = part_end_position, strand = part_strand, color = part_color).exists():
                        plasmid_part = PlasmidPart(name = part_name, start_position = part_start_position, end_position = part_end_position, strand = part_strand, color = part_color)
                        plasmid_part.save()


                        plasmid_parts_writer.add_document(
                            name = part_name,
                            start_position = part_start_position,
                            end_position = part_end_position,
                            strand = part_strand,
                            color = part_color,
                            sql_id = plasmid_part.id
                        )

                        plasmid_part.plasmids.add(plasmid)
                    else:
                        plasmid_parts = PlasmidPart.objects.filter(name = part_name, start_position = part_start_position, end_position = part_end_position, strand = part_strand, color = part_color)

                        for plasmid_part in plasmid_parts:
                            plasmid_part.plasmids.add(plasmid)
            except UnicodeDecodeError:
                print(name_for_fasta)        
        elif len(item) >= 3 and item[-3:] == ".gb":
            try:
                for i, record in enumerate(SeqIO.parse(file_path, "genbank")):
                    plasmid = Plasmid(name = item.split(".")[0], sequence = str(record.seq).upper())
                    plasmid.save()

                    plasmid_writer.add_document(
                        name = item.split(".")[0],
                        sequence = str(record.seq).upper(),
                        sql_id = plasmid.id
                    )

                    for feature in record.features:
                        if "label" in feature.qualifiers:
                            annotation_name = feature.qualifiers["label"][0]
                            annotation_location = feature.location

                            numerical_strand_direction = annotation_location.strand

                            annotation_strand = ""

                            if numerical_strand_direction == -1:
                                annotation_strand = "-"
                            else:
                                annotation_strand = "+"

                            annotation_sequence = annotation_location.extract(record.seq)

                            random_color = "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])

                            while PlasmidPart.objects.filter(color = random_color).exists():
                                random_color = "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])

                            plasmid_part = PlasmidPart(
                                name = annotation_name,
                                start_position = annotation_location.start,
                                end_position = annotation_location.end,
                                strand = annotation_strand,
                                color = random_color
                            )

                            plasmid_part.save()

                            if isinstance(annotation_location, Bio.SeqFeature.CompoundLocation):
                                feature_indices = list(annotation_location)

                                plasmid_parts_writer.add_document(
                                    name = annotation_name,
                                    start_position = feature_indices[0],
                                    end_position = feature_indices[-1],
                                    strand = annotation_strand,
                                    color = random_color,
                                    sql_id = plasmid_part.id
                                )
                            else:
                                plasmid_parts_writer.add_document(
                                    name = annotation_name,
                                    start_position = annotation_location.start,
                                    end_position = annotation_location.end,
                                    strand = annotation_strand,
                                    color = random_color,
                                    sql_id = plasmid_part.id
                                )

                            plasmid_part.plasmids.add(plasmid)
            except UnicodeDecodeError:
                print("Mlem")

    plasmid_writer.commit()
    plasmid_parts_writer.commit()

def search_plasmid_name(query_string):
    plasmid_ix = index.open_dir("Indexed_Search_Database/Plasmid_Name")
    qp = QueryParser("name", schema = plasmid_ix.schema)
    q = qp.parse(query_string)

    searcher = plasmid_ix.searcher()
    results = searcher.search(q)

    return results

def search_plasmid_annotation(query_string):
    plasmid_annotation_ix = index.open_dir("Indexed_Search_Database/Plasmid_Annotation")
    qp = QueryParser("name", schema = plasmid_annotation_ix.schema)
    q = qp.parse(query_string)

    searcher = plasmid_annotation_ix.searcher()
    results = searcher.search(q)

    return results
