from django.shortcuts import render, get_object_or_404
from django.contrib.auth import login, logout
from django.core.mail import send_mail
from django.conf import settings as conf_settings
from django.contrib.auth.models import User, Group

from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response
from rest_framework import generics, permissions
from rest_framework.permissions import IsAuthenticated
from rest_framework.permissions import AllowAny
from rest_framework.authtoken.serializers import AuthTokenSerializer

from rest_framework.parsers import FileUploadParser
from rest_framework.views import APIView

from knox.models import AuthToken
from knox.views import LoginView as KnoxLoginView
from knox import views as knox_views
from ..serializers import UserSerializer, RegisterSerializer
from knox.auth import TokenAuthentication

from django.http import JsonResponse

import iBioFabBackend.cooked_up_libraries.flexible_alignment_search as fas
import iBioFabBackend.cooked_up_libraries.snapgene_annotation_parser as sap
import iBioFabBackend.cooked_up_libraries.api_helper_methods as ahm
import iBioFabBackend.cooked_up_libraries.plasmids_and_parts_updater as ppu

from snapgene_reader import snapgene_file_to_dict
from Bio import SeqIO
from Bio.Seq import Seq

import re
import tempfile

from ..models import Plasmid, PlasmidPart, Order, Feature, UserPersonalInfo, DNAFiles, Notification
from ..documents import PlasmidDocument, PlasmidPartDocument

import random
from io import BytesIO

import json
import os

from django.utils.timesince import timesince

import pandas as pd
import numpy as np

import os
import shutil

# Load information in bulk for user homepage

@api_view(["POST"])
@permission_classes((IsAuthenticated, ))
def get_user_information(request):
    if request.method == "POST":
        user = request.user

        is_technician = request.user.groups.filter(name = "Technician").exists()
        made_request_for_technician_status = request.user.userpersonalinfo.requested_technician_status

        orders = user.order_set.order_by("-created_at")

        response = {
            "basic_info": {
                "first_name": user.userpersonalinfo.first_name,
                "last_name": user.userpersonalinfo.last_name,
                "affiliation": user.userpersonalinfo.affiliation
            },
            "technician_status": {
                "is_technician": is_technician,
                "requested_for_technician_status": made_request_for_technician_status
            }
        }

        order_set = []

        for order in orders:
            file_contents = ahm.get_zip_file_from_order(order)

            if file_contents != False:
                order_set.append({
                    "order_id": order.id,
                    "order_name": order.name,
                    "created_at": order.created_at.strftime("%s"),
                    "updated_at": order.updated_at.strftime("%s"),
                    "status": order.status,
                    "has_file_contents": True,
                    "file_contents": file_contents
                })
            else:
                order_set.append({
                    "order_id": order.id,
                    "order_name": order.name,
                    "created_at": order.created_at.strftime("%s"),
                    "updated_at": order.updated_at.strftime("%s"),
                    "status": order.status,
                    "has_file_contents": False
                })

        response["orders"] = order_set

        return JsonResponse(response)

# API Endpoints regarding loading and updating status of notifications

@api_view(["POST"])
@permission_classes((IsAuthenticated, ))
def load_notifications_for_user(request):
    if request.method == "POST":
        user = request.user

        notifications = user.notification_set.order_by("-created_at")

        response = {
            "non_processed_notifications": notifications.filter(has_read = False).count()
        }

        notifications_list = []

        for notification in notifications:
            notification_dict = {
                "header": notification.header,
                "body": notification.body,
                "response_type": notification.response_type,
                "has_processed": notification.has_read,
                "created_at": order.created_at.strftime("%s"),
                "updated_at": order.updated_at.strftime("%s")
            }

            if "Technician Quote Review" in notification.response_type:
                order_id = int(notification.response.split("|")[1])

                order = Order.objects.get(pk = order_id)

                if order.status == "Quote Sent":
                    notification.has_read = True
                    notification.save()
                    response["non_processed_notifications"] = user.notification_set.filter(has_read = False).count()

                order_sequence = order.sequence
                order_bacterial_selection = order.bacterial_selection
                order_origin_replication = order.originReplication
                order_organism = order.organism

                associated_order = {
                    "order_id": order_id,
                    "name": order.name,
                    "sequence": order_sequence,
                    "backbone": {
                        "bacterial_selection": order_bacterial_selection,
                        "origin_replication": order_origin_replication,
                        "organism": order_organism
                    }
                }

                features = []

                for feature in order.feature_set.all().order_by("start_position"):
                    if feature.has_custom_sequence:
                        features.append({
                            "id": feature.id,
                            "start": feature.start_position,
                            "end": feature.end_position,
                            "name": feature.name,
                            "custom": True,
                        })
                    else:
                        features.append({
                            "id": feature.plasmid_part.id,
                            "start": feature.start_position,
                            "end": feature.end_position,
                            "name": feature.name,
                            "custom": False,
                        })

                associated_order["features"] = features

                notification_dict["order"] = associated_order

            notifications_list.append(notification_dict)

        response["notifications"] = notifications_list

        return JsonResponse(response)

# User registration, login, and logout APIs

# Register API
class RegisterAPI(generics.GenericAPIView):
    serializer_class = RegisterSerializer
    permission_classes = [AllowAny]

    def post(self, request, *args, **kwargs):
        if ahm.has_all_fields_present(request, ["username", "email", "password", "confirm_password", "first_name", "last_name", "affiliation"]):
            if request.data["password"] == request.data["confirm_password"]:
                if request.data["first_name"] != "" and request.data["last_name"] != "" and request.data["affiliation"] != "":
                    serializer = self.get_serializer(data = request.data)
                    serializer.is_valid(raise_exception = True)

                    new_user = serializer.save()

                    token_for_verification = random.getrandbits(256)

                    while UserPersonalInfo.objects.filter(verification_token = token_for_verification).exists():
                        token_for_verification = random.getrandbits(256)

                    user_personal_info = UserPersonalInfo(
                        user = new_user,
                        first_name = request.data["first_name"],
                        last_name = request.data["last_name"],
                        affiliation = request.data["affiliation"],
                        verification_token = token_for_verification,
                        verified = True # stop gap until the verification link works
                    )

                    user_personal_info.save()

                    actual_message_to_put_in_later = "Click on the following link to verify your account and begin ordering custom plasmids.\n\n%sverify/%s" % (conf_settings.FRONTEND_LINK, token_for_verification)

                    send_mail(
                        "Welcome to PlasmidBuilder. Verify account to begin!",
                        actual_message_to_put_in_later,
                        conf_settings.EMAIL_HOST_USER,
                        [new_user.email],
                        fail_silently = False
                    )

                    output_response = {
                        "user": UserSerializer(new_user, context = self.get_serializer_context()).data
                    }

                    return Response(output_response)
                else:
                    errors = []

                    if request.data["first_name"] == "":
                        errors.append("First name field is blank.")

                    if request.data["last_name"] == "":
                        errors.append("Last name field is blank.")

                    if request.data["affiliation"] == "":
                        errors.append("Affiliation field is blank.")

                    return Response({
                        "errors": errors
                    })
            else:
                return Response({
                    "errors": ["Password and confirm password are not the same."]
                })
        else:
            return Response({
                "errors": ["Fields not present in API input"]
            })

# Login API
class LoginAPI(KnoxLoginView):
    permission_classes = (permissions.AllowAny,)

    def post(self, request, format = None):
        serializer = AuthTokenSerializer(data = request.data)
        serializer.is_valid(raise_exception = True)
        user = serializer.validated_data['user']

        if user.userpersonalinfo.verified:
            login(request, user)

            return super(LoginAPI, self).post(request, format = None)
        else:
            return Response({
                "non_field_errors": ["Need to verify account."]
            })

# User Account Verification API
class VerifyAPI(generics.GenericAPIView):
    permission_classes = (permissions.AllowAny,)

    def post(self, request, format = None):
        if ahm.has_all_fields_present(request, ["verification_token"]):
            given_verification_token = request.data["verification_token"]

            if UserPersonalInfo.objects.filter(verification_token = given_verification_token).exists():
                user_personal_info = UserPersonalInfo.objects.get(verification_token = given_verification_token)

                user_personal_info.verified = True
                user_personal_info.save()

                user = user_personal_info.user

                return Response({
                    "message": "You are sucessfully verified!",
                    "token": AuthToken.objects.create(user)[1]
                })
            else:
                return Response({
                    "message": "Verification token invalid."
                })
        else:
            return Response({
                "message": "Missing fields present."
            })

### API calls for when user forgets password

@api_view(["POST"])
@permission_classes((AllowAny, ))
def request_password_change(request):
    if request.method == "POST":
        if ahm.has_all_fields_present(request, ["email"]):
            try:
                user = User.objects.get(email = request.data["email"])

                token_for_verification = random.getrandbits(256)

                while UserPersonalInfo.objects.filter(verification_token = token_for_verification).exists():
                    token_for_verification = random.getrandbits(256)

                user_personal_info = user.userpersonalinfo

                user_personal_info.verification_token = token_for_verification
                user_personal_info.save()

                send_mail(
                    "PlasmidBuilder Password Reset Instructions",
                    "Hello %s %s,\n\nClick on the following link below to reset your password.\n\n%spassword_reset_form/%s" % (user_personal_info.first_name, user_personal_info.last_name, conf_settings.FRONTEND_LINK, user_personal_info.verification_token),
                    conf_settings.EMAIL_HOST_USER,
                    [user.email],
                    fail_silently = False
                )

                return Response({
                    "success": "Email successfuly sent"
                })
            except Exception as err:
                print(err)
                return Response({
                    "error": "Email does not exist"
                })
        else:
            return Response({
                "error": "Missing fields present"
            })

@api_view(["POST"])
@permission_classes((AllowAny, ))
def submit_password_change(request):
    if request.method == "POST":
        if ahm.has_all_fields_present(request, ["password", "confirm_password", "verification_token"]):
            if request.data["password"] == request.data["confirm_password"]:
                try:
                    user_personal_info = UserPersonalInfo.objects.get(verification_token = request.data["verification_token"])

                    user = user_personal_info.user
                    user.set_password(request.data["password"])
                    user.save()

                    return Response({
                        "success": "Password successfully changed"
                    })
                except Exception as err:
                    print(err)

                    return Response({
                        "error": "User does not exist"
                    })
            else:
                return Response({
                    "error": "Password and Confirm Password Not the Same"
                })
        else:
            return Response({
                "error": "Missing fields present"
            })